/*
 * (c) Copyright 2001-2003
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#ifdef _WIN32
#include <conio.h>
#else
#include <sys/time.h>
#endif

#include "s68.h"
#include "m68000.h"
#include "decoder.h"
#include "conf.h"
#include "ui.h"

#include "s68stat.h"


#define CHKCOUNT        1000    /* Check input after 1000 instr. */

int   bpoints = 0;              /* Number of breakpoints */
long  brkpt[100];               /* Breakpoints */

static char    *args[16];       /* Pointer to command line words */
static int     nextCmd;         /* Index to command from conf file */
static char    lbuf[128];       /* Command line buffer */
static char    olbuf[128];      /* Old Command line buffer */



/*
 * c h k I n p u t
 *
 * Check for pressed key
 */
static int chkInput(void)
{
#ifndef _WIN32
    struct timeval timeout;
    fd_set inFdSet;

    FD_ZERO(&inFdSet); FD_SET(0, &inFdSet);
    timeout.tv_sec = timeout.tv_usec = 0;
    select(0+1, &inFdSet, NULL, NULL, &timeout);
    if (FD_ISSET(0, &inFdSet))  {
        getchar();
        return 1;
    }
#endif

    return 0;
}



/*
 * h e C m d
 *
 * Print help
 */
static void heCmd(int argc, char *argv[])
{
    char *cmd;

    if (argc == 1)  {
        puts("\nSimulator Commands:\n");
        puts(" he [<cmd>]               Print detailed help about command <cmd>");
        puts(" ver                      Display version of simulator");
        puts(" ex                       Exit MC68000 simulator");
        puts(" br [<addr>]              Display or Set breakpoint");
        puts(" nb [<addr>]              Clear all or one breakpoint");
        puts(" md [<addr>] [<addr>]     Memory display");
        puts(" mf <addr> <addr> <val>   Memory fill");
        puts(" mm [<addr>]              Memory modify");
        puts(" rd                       Register display");
        puts(" rm <reg> <val>           Register modify");
        puts(" go [<cnt>]               Execute <cnt> instructions");
        puts(" tr [<cnt>]               Execute <cnt> instructions with tracing");
        puts("");

        return;
    }

    cmd = argv[1];

    if (!strcmp(cmd, "he"))  {
        printf(" Syntax: he [<cmd>]\n");
        printf(" Description: Print detailed description of command\n");
    }
    else if (!strcmp(cmd, "go"))  {
        puts(" Syntax: go [<cnt>]");
        puts(" Description: Execute <cnt> instructions");
        puts("              Without <cnt> execute endless");
    }
    else if (!strcmp(cmd, "tr"))  {
        puts(" Syntax: tr [<cnt>]");
        puts(" Description: Execute <cnt> instructions with disassembling");
        puts("              Without <cnt> execute endless");
    }
    else if (!strcmp(cmd, "mm"))  {
        puts(" Syntax: mm <addr>");
        puts(" Description: Modify memory at <addr>");
        puts("              Use '.' to exit the memory modify mode");
        puts("              and '-' to go to previouse address");
    }
    else if (!strcmp(cmd, "rm"))  {
        puts(" Syntax: rm <reg> <val>");
        puts(" Description: Modify register <reg>");
        puts("              Valid registers are d0 to d7, a0 to a8, and pc");
        puts("              a7 is the usp and a8 is used for the ssp");
    }
    else  {
        fprintf(stderr, "ERROR: No description for '%s'\n", cmd);
    }
}



/*
 * e x C m d
 *
 * Exit Simulator
 */
static void exCmd(int argc, char *argv[])
{
    exit(0);
}



/*
 * v e C m d
 *
 * Display program version 
 */
static void veCmd(int argc, char *argv[])
{
    printf("Version %s [%s]\n", S68VER, __DATE__);
}



/*
 * r d C m d
 *
 * Display processor registers
 */
static void rdCmd(int argc, char *argv[])
{
    int   j;
    char  srbits[] = "T S  INT   XNZVC";

    printf("Dn:");
    for (j = 0; j <= 7; j++)  {
        if (j == 4)  printf(" ");
        printf(" %08lx",D[j]);
    }
    printf("\n");

    printf("An:");
    for (j = 0; j <= 7; j++)  {
        if (j == 4)  printf(" ");
        printf(" %08lx",A[j]);
    }
    printf("\n");

    printf("PC: %08lx      ", PC);

    printf("SR: %04x  ", (unsigned short) SR);
    for (j = 15; j >= 0; j--)  {
        if ((0x1 << j) & SR)    /* print each bit of SR */
            printf ("%c", srbits[15 - j]);
        else
            printf (".");
    }

    printf("                        %08lx\n", A[8]);
}



/*
 * m d C m d
 *
 * Memory display
 */
static void mdCmd(int argc, char *argv[])
{
    u_long addr, eaddr;
    u_long val;
    int   i;

    if (argc > 1)  {
        if (sscanf(argv[1], "%lx", &addr) != 1)  {
            fprintf(stderr, "ERROR: Invalid address\n");
            return;
        }
    }
    else  {
        addr = 0x000000;
    }

    if (argc < 3)  {
        eaddr = addr + 0x0100;
    }
    else  {
        if (sscanf(argv[2], "%lx", &eaddr) != 1)  {
            fprintf(stderr, "ERROR: Invalid end address\n");
            return;
        }
    }

    while (1) {
        for (i = 0; i < 16; i++)  {
            if (memRead(addr, SZ_BYTE, &val) != SUCCESS)  {
                fprintf(stderr, "ERROR: Bus Error at %06lx\n", addr);
                return;
            }
            if (i == 0)  printf("%06lx : ", addr);
            printf(" %02x", (unsigned char)(val & 0xff));
            if (++addr >= eaddr)  {
                printf("\n");
                return;
            }
        }
        printf("\n");
    }
}



/*
 * m f C m d
 *
 * Memory fill
 */
static void mfCmd(int argc, char *argv[])
{
    u_long addr, eaddr;
    u_long val;

    if (argc < 3)  {
        fprintf(stderr, "ERROR: Syntax\n");
        return;
    }

    if (sscanf(argv[1], "%lx", &addr) != 1)  {
        fprintf(stderr, "ERROR: Invalid address\n");
        return;
    }
    if (sscanf(argv[2], "%lx", &eaddr) != 1)  {
        fprintf(stderr, "ERROR: Invalid end address\n");
        return;
    }
    if (sscanf(argv[3], "%lx", &val) != 1)  {
        fprintf(stderr, "ERROR: Invalid value\n");
        return;
    }

    for ( ; addr < eaddr; addr++)  {
        if (memWrite(addr, SZ_BYTE, val) != SUCCESS)  {
            fprintf(stderr, "ERROR: Bus Error at %06lx\n", addr);
            return;
        }
    }
}



/*
 * m m C m d
 *
 * Memory Modify   
 */
static void mmCmd(int argc, char *argv[])
{
    u_long  addr;
    u_long  val, newVal;
    char    mmBuf[128];

    if (argc > 1)  {
        if (sscanf(argv[1], "%8lx", &addr) != 1)  {
            fprintf(stderr, "ERROR: Illegal Address\n");
            return;
        }
    }
    else {
        addr = 0x000000;
    }

    while (1)  {
        if (memRead(addr, SZ_BYTE, &val))  {
            fprintf(stderr, "ERROR: Bus Error at %06lx\n", addr);
            return;
        }
        printf("%06lx = %02x   ? ", addr, (unsigned char) val);
        if (fgets(mmBuf, 80, stdin) == NULL)  exit(0);
        if (mmBuf[0] == '.')  break;
        if (mmBuf[0] == '\n')  {
            addr++;
            continue;
        }
        if (mmBuf[0] == '-')  {
            addr--;
            continue;
        }
        if (sscanf(mmBuf, "%lx", &newVal) != 1)  {
            fprintf(stderr, "ERROR: Illegal Value\n");
            return;
        }
        if (newVal > 0xFF)  {
            fprintf(stderr, "ERROR: Value out of range\n");
            return;
        }
        if (memWrite(addr, SZ_BYTE, newVal))  {
            fprintf(stderr, "ERROR: Bus Error at %06lx\n", addr);
            return;
        }
        else  {
            addr++;
        }
    }
}



/*
 * r m C m d
 *
 * Register Modify
 */
static void rmCmd(int argc, char *argv[])
{
    u_long newVal;
    int    regNum;

    if (argc < 3)  {
        fprintf(stderr, "ERROR: Syntax\n");
        return;
    }
    if (sscanf(argv[2], "%lx", &newVal) != 1)  {
        fprintf(stderr, "ERROR: Invalid number\n");
        return;
    }

    if (!strcmp(argv[1], "pc"))  {
        PC = newVal;
    }
    else if (!strcmp(argv[1], "ssp"))  {
        SSP = newVal;
    }
    else if (!strcmp(argv[1], "usp"))  {
        USP = newVal;
    }
    else if (argv[1][0] == 'd')  {
        if (sscanf(&argv[1][1], "%d", &regNum) != 1)  {
            fprintf(stderr, "ERROR: Invalid register number\n");
            return;
        }
        if ((regNum < 0) || (regNum > 7))  {
            fprintf(stderr, "ERROR: Register number out of range\n");
            return;
        }
        D[regNum] = newVal;
    }
    else if (argv[1][0] == 'a')  {
        if (sscanf(&argv[1][1], "%d", &regNum) != 1)  {
            fprintf(stderr, "ERROR: Invalid register number\n");
            return;
        }
        if ((regNum < 0) || (regNum > 8))  {
            fprintf(stderr, "ERROR: Register number out of range\n");
            return;
        }
        A[regNum] = newVal;
    }
    else  {
        fprintf(stderr, "ERROR: Unkown register\n");
    }
}



/*
 * r e a d L i n e
 *
 * Read one line from S-Record file
 */
static char *readLine(char *bp, int sz, FILE *f)
{
    int  c, i;

    for (i = 0; i < sz; i++)  {
        c = getc(f);
        if ((c == '\r') || (c == '\n')) {
            bp[i] = '\0';
            break;
        }
        bp[i] = (char) c;
    }

    return bp;
}



/*
 * l d C m d
 *
 * Load S-Record to memory
 */
static void ldCmd(int argc, char *argv[])
{
    FILE  *fp;
    char  *fileName;
    char  buffer[128];
    char  *bptr, type;
    int   line = 0;
    u_long i;
    u_long address, nbyte, data;

    if (argc == 0)  {
        fprintf(stderr, "ERROR: Missing filename\n");
        return;
    }

    fileName = argv[1];

    if ((fp = fopen(fileName, "r")) == NULL)  {
        fprintf(stderr, "ERROR: Can't open %s\n", fileName);
        return;
    }

    printf(" -------- "); fflush(stdout);

    while (!feof(fp))  {
        line++;
        if ((bptr = readLine(buffer, 128, fp)) == NULL)  {
            fprintf(stderr, "ERROR: Unexpected end of file - %s\n", strerror(errno));
            return;
        }

        /* Get Type */
        if (sscanf(bptr, "S%c", &type) != 1)  {
            /* Ignore lines that don't start with S */
            break;
        }
        bptr += 2;

        /* Get Length */
        if (sscanf(bptr, "%02lx", &nbyte) != 1)  {
            fprintf(stderr, "ERROR: Line %d: Bad character in length\n", line);
            return;
        }
        bptr += 2;

        /* Get address field */
        switch (type)  {
            case '0' :
                /* Ignore start record */
	        continue;

            case '1' :
                if (sscanf(bptr,"%04lx", &address) != 1)  {
                    fprintf(stderr, "ERROR: Line %d: Bad address\n", line);
                    return;
                }
                bptr += 4;
                nbyte -= 2;
                break;

            case '2' :
                if (sscanf(bptr,"%06lx", &address) != 1)  {
                    fprintf(stderr, "ERROR: Line %d: Bad address\n", line);
                    return;
                }
                bptr += 6;
                nbyte -= 3;
                break;

            case '3' :
                if (sscanf(bptr,"%08lx", &address) != 1)  {
                    fprintf(stderr, "ERROR: Line %d: Bad address\n", line);
                    return;
                }
                bptr += 8;
                nbyte -= 4;
                break;

            case '7' :
            case '8' :
            case '9' :
                /* Ignore all termination records */
                continue;

            default :
                fprintf(stderr, "ERROR: Line %d: Bad type '%c'\n", line, type);
                return;
        }
        printf(" \b\b\b\b\b\b\b\b\b\b\b%08lx ", address);
        fflush(stdout);

        /* Get all data bytes without checksum */
        for (i = 0; i < nbyte-1; i++)  {
            if (sscanf(bptr, "%02lx", &data) != 1)  {
                fprintf(stderr, "ERROR: Line %d: Bad character in data\n", line);
                return;
            }
            bptr += 2;
            if (memWrite(address++, SZ_BYTE, data) != 0)  {
                fprintf(stderr, "ERROR: Bus Error at 0x%08lx\n", address-1);
                return;
            }
        }

        /* Ignore checksum */
        if (sscanf(bptr, "%02lx", &data) != 1)  {
            fprintf(stderr, "ERROR: Line %d: Bad character in checksum\n", line);
            return;
        }
    }
    printf("\n");

    fclose(fp);

}



/*
 * b r C m d
 *
 * Display or Set breakpoint
 */
static void brCmd(int argc, char *argv[])
{
    int   i;
    long  addr;

    if (argc > 1)  {
        /* Set Breakpoint */
        if (sscanf(argv[1], "%lx", &addr) != 1)  {
            fprintf(stderr, "ERROR: Illegal Address\n");
            return;
        }

        for (i = 0; i < bpoints; i++)
            if (brkpt[i] == addr)  return;
            brkpt[bpoints++] = addr;
            if (bpoints > 100)
                bpoints = 100;
    }
    else  {
        /* Display breakpoints */
        for (i = 0; i < bpoints; i++)
            printf("#%2d   %08lx\n", i+1, brkpt[i]);
    }
}



/*
 * n b C m d
 *
 * Clear breakpoint
 */
static void nbCmd(int argc, char *argv[])
{
    long   addr;
    int    i, j;

    if (argc > 1)  {
        /* Clear one Breakpoint */
        if (sscanf(argv[1], "%lx", &addr) != 1)  {
            fprintf(stderr, "ERROR: Illegal breakpoint address\n");
            return;
        }

        /* Find breakpoint */
        for (i = 0; i < bpoints; i++)
            if (brkpt[i] == addr)  break;

        if (i == bpoints)  {
            fprintf(stderr, "ERROR: No break point at %08lx\n", addr);
            return;
        }
        --bpoints;

        /* Adjust breakpoint table and */
        for (j = i; j < bpoints; j++)  {
            brkpt[j] = brkpt[j + 1];
        }
    }
    else  {
        /* Clear all breakpoints */
        bpoints = 0;
    }
}



/*
 * g o C m d
 *
 * Start simulation
 */
static void goCmd(int argc, char *argv[])
{
    int   i;
    long  chkCount = CHKCOUNT;
    long  count = 0;	/* Default is endless */

    if (argc > 1)  {
        if (sscanf(argv[1], "%lx", &count) != 1)  {
            fprintf(stderr, "ERROR: Illegal counter\n");
            return;
        }
    }

    if (!strcmp("tr", argv[0]))  trace = 1;

    tStart();

    while (1)  {
        instCounter++;
        cpuExec();

        /* Check for any breakpoints */
        for (i = 0; i < bpoints; i++)  {
            /* if PC equals a breakpoint, break */
            if (PC == brkpt[i])  {
                printf("BREAK POINT at %08lx\n", PC);
                trace = 0;
                return;
            }
        }

        /* Check if ENTER key is pressed */
        if (--chkCount == 0)  {
            chkCount = CHKCOUNT;
            if (chkInput())  {
                printf("STOPPED at %08lx\n", PC);
                trace = 0;
                tStop();
                tReport();
                return;
            }
        }

        /* Check if counter is zero */
        if (count && (--count == 0))  {
            trace = 0;
            return;
        }
    }

    trace = 0;
}



/*
 * r e a d C m d L i n e
 *
 * Read one command line
 */
static char *readCmdLine(void)
{
    char  *cl;

    if (nextCmd < conf.numCmd)  {
        cl = conf.cmd[nextCmd++];
        if (!quiet)  printf("%s\n", cl);
    }
    else  {
        cl = fgets(lbuf, 80, stdin);
        if (lbuf[0] == '\n')
            strcpy(lbuf, olbuf);
        else
            strcpy(olbuf, lbuf);
    }

    return cl;
}



/*
 * g e t A r g s
 *
 * Get arguments of command line
 */
static int getArgs(char *str, char *argPtr[], int maxCnt)
{
    int  argCnt = 0;

    while (*str)  {
        while (isspace(*str))  str++;           /* Skip leading white space */
        if (*str == '#') break;                 /* Comment introducer */
        if (*str)  {
            argPtr[argCnt++] = str;
            while (!isspace(*str))
                if (!*str++) return(argCnt);
            *str++ = '\0';
        }
        if (argCnt == maxCnt) break;            /* Enough args scanned */
    }

    return argCnt;
}



/*
 * r s t C m d
 *
 * Reset CPU and Decoder
 */
static void rstCmd(int argc, char *argv[])
{
    decoderReset();
    cpuReset();
}



/*
 * One entry in the command table
 */
typedef struct  {
    char  *name;
    void  (*func)(int argc, char *argv[]);
} oneCmd;

/*
 * c m d T a b l e
 *
 * Command table
 */
static oneCmd  cmdTable[] = {
    {"?",    heCmd},     /* Print help info */
    {"he",   heCmd},     /* Print help info */
    {"help", heCmd},     /* Print help info */
    {"ver",  veCmd},     /* Display version */
    {"ex",   exCmd},     /* End of simulation fun */
    {"exit", exCmd},     /* End of simulation fun */
    {"br",   brCmd},     /* Display or Set Breakpoints */
    {"nb",   nbCmd},     /* Clear all or one Breakpoints */
    {"rd",   rdCmd},     /* Display CPU registers */
    {"rm",   rmCmd},     /* Register modify */
    {"md",   mdCmd},     /* Memory display */
    {"mf",   mfCmd},     /* Memory fill */
    {"mm",   mmCmd},     /* Memory modify */
    {"ld",   ldCmd},     /* Load S-Record from file */
    {"go",   goCmd},     /* Start simulation */
    {"tr",   goCmd},     /* Start simulation with disassembling */
    {"rst",  rstCmd},    /* Reset the system */
    {NULL,   NULL}       /* End of command table */
};



/*
 * u i S t a r t
 *
 * Start the command line user interface
 */
int uiStart(void)
{
    char  *cmdLine;
    oneCmd *cp;
    int   argc = 0;
    char  **argv = args;

    nextCmd = 0;

    /* Read commands and execute */
    while (1)  {
        if (!quiet)  printf("s68> ");
        if ((cmdLine = readCmdLine()) == NULL) 
            exit(0);
        argc = getArgs(cmdLine, argv, 10);

        if (argc == 0)  continue;

        /* Search in command table */
        for (cp = cmdTable; cp->name != NULL; cp++)  {
            if (!strcmp(cp->name,  argv[0]))  {
                cp->func(argc, argv);
                break;
            }
        }
        if (cp->name == NULL) {
            fprintf(stderr, "ERROR: Unknown command. Try 'he' for help\n");
        }
    }

    return 0;
}

