/*
 * (c) Copyright 2001-2003
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <stdio.h>
#include <stdlib.h>

#include "s68.h"
#include "ui.h"
#include "m68000.h"
#include "decoder.h"
#include "opcodes.h"
#include "conf.h"


int   trace;                    /* Enable disassembling */
long  D[8], A[9], PC, SR;       /* Processor Register */
int   cpuState;                 /* CPU state */
long  inst;                     /* Current instruction */

static int   intLevel;          /* Current interrupt level pending */
static int   cpuCounter;

static void  exception(int vector, int class, long addr, int r_w);



/*
 * c p u R e s e t
 *
 * Reset the M68000
 */
int cpuReset(void)
{
    u_long  tmpPC, tmpA7;

    SR = 0x2700;
    cpuState = ST_RESET;

    if (memRead(0x000000, SZ_LONG, &tmpA7) || memRead(0x000004, SZ_LONG, &tmpPC))  {
        uiErrPrintf("M68000: Bus Error during Reset vector read\n");
        cpuState = ST_HALT;
        return 0;
    }
    A[8] = tmpA7;
    PC = tmpPC;
    cpuState = ST_RUNNING;
 
    return 0;
}


/*
 * c p u I n i t
 *
 * Initialize the M68000
 */
int cpuInit(void)
{
    int   i;

    for (i = 0; i <= 7; i++)
        A[i] = D[i] = 0x00000000;
    A[8] = 0x00000000;
    PC = 0x00000000;
    SR = 0x2700;

    intLevel = 0;                      /* No pending interrupt */
    cpuCounter = conf.numInstCycle;    /* Preset polling counter */

    return 0;
}


/*
 * m e m P u t
 *
 * Put data to memory
 * This function also performs the exception processing.
 */
int memPut(long data, long addr, long size)
{
    if ((addr & 0x000001) && (size != SZ_BYTE))  {
        exception(3, 0, addr, WRITE);
        return ADDRESS_ERROR;
    }

    if (memWrite(addr, size, data))  {
        exception(2, 0, addr, WRITE);
        return BUS_ERROR;
    }

    return SUCCESS;
}


/*
 * m e m G e t
 *
 * Get data from memory
 * This function also performs the exception handling
 */
int memGet(long addr, long size, long *result)
{
    u_long  temp;

    if ((addr & 0x000001) && (size != SZ_BYTE))  {
        exception(3, 0, addr, READ);
        return ADDRESS_ERROR;
    }

    if (memRead(addr, size, &temp))  {
        exception(2, 0, addr, READ);
        return BUS_ERROR;
    }

    *result = (long)temp;

    return SUCCESS;
}


/*
 * m e m R e q u e s t
 *
 * Get data from code area
 */
int memRequest(long *addrPtr, long size, long *result)
{
    u_long  addr = *addrPtr;
    u_long  temp;

    if ((addr & 0x000001) && (size != SZ_BYTE))  {
        exception(3, 0, addr, READ);
        return ADDRESS_ERROR;
    }

    if (memRead(addr, (size == SZ_BYTE) ? SZ_WORD : size, &temp))  {
        exception(2, 0, addr, READ);
        return BUS_ERROR;
    }

    *result = temp;

    if (size == SZ_LONG)  {
        addr += 4;
    }
    else if (size == SZ_WORD) {
        addr += 2;
    }
    else  {
        *result &= 0x00ff;
        addr += 2;
    }
    *addrPtr = addr;

    return SUCCESS;
}



/*
 * e x c e p t i o n
 *
 * Exception handling
 */
static void exception(int vector, int class, long addr, int r_w)
{
    int   iword;
    long  tmpSR;

    tmpSR = SR;

    SR = SR | sbit;     /* force processor into supervisor state */
    SR = SR & ~tbit;    /* turn off trace mode */

    if ( (class == 1) || (class == 2))  {
        SSP -= 4; memPut(PC, SSP, SZ_LONG);
        SSP -= 2; memPut(tmpSR, SSP, SZ_WORD);
    }
    else {
        iword = (r_w == READ) ? 0x0036 : 0x0006;
        SSP -= 4;
        if (SSP & 0x00000001)           { cpuState = ST_HALT; return; }
        if (memWrite(SSP, SZ_LONG, PC))    { cpuState = ST_HALT; return; }
        SSP -= 2;
        if (memWrite(SSP, SZ_WORD, tmpSR)) { cpuState = ST_HALT; return; }
        SSP -= 2;
        if (memWrite(SSP, SZ_WORD, inst))  { cpuState = ST_HALT; return; }
        SSP -= 4;
        if (memWrite(SSP, SZ_LONG, addr))  { cpuState = ST_HALT; return; }
        SSP -= 2;
        if (memWrite(SSP, SZ_WORD, iword)) { cpuState = ST_HALT; return; }
    }

    memGet(vector << 2, SZ_LONG, &PC);
}



/*
 * c p u E x e c
 *
 * Executes a single instruction
 */
void cpuExec(void)
{
    int  vector;
    int  funcRet;
    int  ret;

    if (--cpuCounter == 0)  {
        /* Call decoders cycle function */
        decCycle();
        cpuCounter = conf.numInstCycle;
    }

    if (cpuState == ST_HALT)  {
        TrcPrintf(("CPU Halted\n"));
        return;
    }

    if (intLevel && (intLevel > (int)((SR & 0x0700) >> 8)))  {
        vector = intAcknowledge(intLevel);
        TrcPrintf(("Interrupt vector %02x\n", vector));
        exception(vector, 1, 0, READ);
        SR = ((SR & ~0x0700) | (intLevel << 8));
        cpuState = ST_RUNNING;
        return;
    }

    if (cpuState == ST_STOP)  {
        TrcPrintf(("CPU Stopped\n"));
        return;
    }

    /* Get opcode from memory and execute */
    if ((ret = memGet(PC, SZ_WORD, &inst)) != SUCCESS)  {
        /* Bus or Address error */
        return;
    }

    TrcPrintf(("%08lx: %04lx", PC, inst));
    TrcPrintf(("\t%s", (instTable[inst].disasm == NULL) ? "" : instTable[inst].disasm));

    /* Execute instruction */
    PC += 2;
    funcRet = instTable[inst].func();

    TrcPrintf(("\n"));

    switch (funcRet)  {
        case SUCCESS  :
        case SUCCESS_PRIV  :
            break;

        case ILLEGAL_INST :
            TrcPrintf(("ILLEGAL Instruction at %08lx\n", PC-2));
            exception(4, 1, 0, READ);
            break;

        case DIV_BY_ZERO :
            TrcPrintf(("Zero Divide at %08lx\n", PC-2));
            exception(5, 2, 0, READ);
            break;

        case CHK_EXCEPTION :
            TrcPrintf(("CHK Instruction at %08lx\n", PC-2));
            exception(6, 2, 0, READ);
            break;

        case TRAPV_TRAP :
            TrcPrintf(("TRAPV Instruction at %08lx\n", PC-2));
            exception(7, 2, 0, READ);
            break;

        case NO_PRIVILEGE :
            TrcPrintf(("Privilege Violation at %08lx\n", PC-2));
            exception(8, 1, 0, READ);
            break;

        case LINE_1010 :
            TrcPrintf(("Line A trap at %08lx\n", PC-2));
            exception(10, 1, 0, READ);
            break;

        case LINE_1111 :
            TrcPrintf(("Line F trap at %08lx\n", PC-2));
            exception(11, 1, 0, READ);
            break;

        case TRAP_TRAP :
            TrcPrintf(("TRAP Instruction at %08lx\n", PC-2));
            exception(32 + (inst & 0x000F), 2, 0, READ);
            break;
    }

    if ((funcRet != SUCCESS_PRIV) && (SR & tbit))  {
        TrcPrintf(("Trace Trap at %08lx\n", PC-2));
        exception(9, 1, 0, READ);
    }
}



/*
 * c p u S e t I P L
 *
 * Request interrupt
 */
void cpuSetIPL(int level)
{
	intLevel = level;
}

