/*
 * (c) Copyright 2001-2003
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>

#include "s68.h"
#include "device.h"
#include "decoder.h"
#include "conf.h"


#define  REGSIZE   32

#define  TCR	0x01	/* Timer Control Register */
#define  TIVR	0x03	/* Timer Interrupt Vector */
#define  NULR	0x05
#define  CPRH	0x07	/* Counter Preload high */
#define  CPRM	0x09	/* Counter Preload middle */
#define  CPRL	0x0b	/* Counter Preload low */
#define  NULTR	0x0d
#define  CNTRH	0x0f	/* Counter Register high */
#define  CNTRM	0x11	/* Counter Register middle */
#define  CNTRL	0x13	/* Counter Register low */
#define  TSR	0x15	/* Timer Status Register */


typedef struct  {
    u_char  tcr;
    u_char  tivr;
    u_char  cprh, cprm, cprl;
    u_char  cntrh, cntrm, cntrl;
    u_char  tsr;
} M68230;



/*
 * s e t I n t e r r u p t
 *
 * Request Interrupt
 */
static void setInterrupt(Device *dev)
{
    M68230  *m68230 = (M68230 *) dev->priv;

    if (m68230->tsr && (m68230->tcr & 0xE0))
        intSetIRQ(dev);
    else
        intClearIRQ(dev);
}



/*
 * m68230Reset
 *
 * Reset the device
 */
int m68230Reset(Device *dev)
{
    M68230  *m68230 = (M68230 *) dev->priv;

    if (!quiet)  uiPrintf(" M68230");

    m68230->tivr = 0x0F;    /* Uninitialized interrupt vector */
    m68230->tsr = 0x00;     /* Clear m68230 status and mask register */

    setInterrupt(dev);

    return 0;
}



/*
 * m 6 8 2 3 0 R e a d
 *
 * Read one byte
 */
static u_long m68230Read(Device *dev, u_long offset)
{
    M68230  *m68230 = (M68230 *) dev->priv;
    u_char data;

    switch (offset & (REGSIZE-1))  {
        case TCR :
            data = m68230->tcr;
            break;
        case TIVR :
            data = m68230->tivr;
            break;

        case NULR :
            data = 0x00;
            break;
        case CPRH :
            data = m68230->cprh;
            break;
        case CPRM :
            data = m68230->cprm;
            break;
        case CPRL :
            data = m68230->cprl;
            break;

        case NULTR :
            data = 0x00;
            break;
        case CNTRH :
            data = m68230->cntrh;
            break;
        case CNTRM :
            data = m68230->cntrm;
            break;
        case CNTRL :
            data = m68230->cntrl;
            break;

        case TSR :
            data = m68230->tsr;
            break;

        default :
            data = 0xff;
            break;
    }

    return (u_long) data;
}



/*
 * m 6 8 2 3 0 W r i t e
 *
 * Write one byte
 */
static void m68230Write(Device *dev, u_long offset, u_char data)
{
    M68230  *m68230 = (M68230 *) dev->priv;

    switch (offset & (REGSIZE-1))  {
        case TCR :
            m68230->tcr = data;
            if (data & 0x01)  {
                m68230->cntrh = m68230->cprh;
                m68230->cntrm = m68230->cprm;
                m68230->cntrl = m68230->cprl;
            }
            break;

        case TIVR :
            m68230->tivr = data;
            break;

        case NULR :
            break;
        case CPRH :
            m68230->cprh = data;
            break;
        case CPRM :
            m68230->cprm = data;
            break;
        case CPRL :
            m68230->cprl = data;
            break;

        case NULTR :
            break;
        case CNTRH :
            m68230->cntrh = data;
            break;
        case CNTRM :
            m68230->cntrm = data;
            break;
        case CNTRL :
            m68230->cntrl = data;
            break;

        case TSR :
            m68230->tsr = data;
            break;
    }
}



/*
 * m 6 8 2 3 0 I A c k
 *
 * Interrupt Acknowledge
 */
static u_char m68230IAck(Device *dev, int level)
{
    M68230  *m68230 = (M68230 *) dev->priv;

    intClearIRQ(dev);

    return m68230->tivr;
}



/*
 * m 6 8 2 3 0 C y c l e
 *
 * Cycle function
 */
static void m68230Cycle(Device *dev)
{
    M68230  *m68230 = (M68230 *) dev->priv;
    long   cntr;

    if (m68230->tcr & 0x01)  {
        /* Timer is enabled */
        cntr = (m68230->cntrh << 16) + (m68230->cntrm << 8) + m68230->cntrl;

        /* Decrement counter */
        if ((cntr -= 100) < 0x000000)  cntr = 0x000000;

        /* Check if zero */
        if (cntr == 0x000000)  {
            m68230->tcr &= 0xFE;
            m68230->tsr |= 0x01;
            setInterrupt(dev);
        }

        /* Move counter to register */
        m68230->cntrh = (cntr >> 16) & 0xFF;
        m68230->cntrm = (cntr >> 8) & 0xFF;
        m68230->cntrl = (cntr >> 0) & 0xFF;
    }
}



/*
 * i n i t M 6 8 2 3 0
 *
 * Initialize
 */
int initM68230(Device *dev, u_long base, u_long size, void *arg)
{
    M68230  *m68230;

    if ((m68230 = malloc(sizeof(M68230))) == NULL) {
        uiErrPrintf("M68230: Malloc of %ld bytes failed - %s\n", size, strerror(errno));
        return(errno);
    }

    dev->base = (u_long) base;
    dev->mask = ~(size-1);
    dev->priv = (void *) m68230;
    dev->devReset = m68230Reset;
    dev->devRead  = m68230Read;
    dev->devWrite = m68230Write;
    dev->devIAck  = m68230IAck;
    dev->devCycle = m68230Cycle;

    return 0;
}
