/*
 * (c) Copyright 2001-2003
 * Peter Blum, Peter.Blum@readme1st.de
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 * Version: 2.1.0.0 - Aug 25 2003 11:11:11
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#ifdef WIN32
#include <conio.h>
#else
#include <unistd.h>
#include <sys/time.h>
#include <termios.h>
#endif

#include "s68.h"
#include "device.h"
#include "decoder.h"
#include "conf.h"


#ifndef WIN32
struct termios idevSettings;
struct termios devSettings;
#endif

#define  REGSIZE   32

#define  MRxA	0x01	/* Mode Register A */
#define  SRA	0x03	/* Status Register A */
#define  CSRA	0x03	/* Clock Select Regiser A */
#define  CRA	0x05	/* Command Regiser A */
#define  RBA	0x07	/* Receive Buffer A */
#define  TBA	0x07	/* Transmit Buffer A */
#define  ISR	0x0B	/* Interrupt Status Register */
#define  IMR	0x0B	/* Interrupt Mask Register */

#define  MRxB	0x11	/* Mode Register B */
#define  SRB	0x13	/* Status Register B */
#define  CSRB	0x13	/* CLock Select Register */
#define  CRB	0x15	/* Command Register */
#define  RBB	0x17	/* Receive Buffer B */
#define  TBB	0x17	/* Transmit Buffer B */
#define  IVR	0x19	/* Interrupt Vector Register */

/* Bits in status register */
#define  RxRDY  0x01
#define  FFULL  0x02
#define  TxRDY  0x04
#define  TxEMT  0x08

/* Channel state */
#define  ACTIVE		1
#define  INACTIVE	0


typedef struct  {
	u_char  mr1a, mr2a;
	u_char  sra, csra;
	u_char  cra;
	u_char  rba, tba;
	u_char  ipcr;

	u_char  acr;
	u_char  isr;
	u_char  imr;
	u_char  cur;
	u_char  cturr;
	u_char  clr;
	u_char  ctlr;

	u_char  mr1b;
	u_char	mr2b;
	u_char  srb, csrb;
	u_char  crb;
	u_char  rbb, tbb;

	u_char  ivr;

	int	mr1aPtr, mr1bPtr;
	int	rxaState, txaState;
	int	rxbState, txbState;
	int     aFd, bFd;		/* FD number of port A and B file */
} M68681;




/*
 * s e t I n t e r r u p t
 *
 * Request Interrupt
 */
static void setInterrupt(Device *dev)
{
    M68681  *m68681 = (M68681 *) dev->priv;
    u_char  tmp_isr = 0x00;

    /* Check Port A */
    if (m68681->sra & TxRDY)  tmp_isr |= 0x01;
    if (m68681->mr1a & 64)  {
        if (m68681->sra & FFULL)  tmp_isr |= 0x02;
    }
    else  {
        if (m68681->sra & RxRDY)  tmp_isr |= 0x02;
    }

    /* Check Port B */
    if (m68681->srb & TxRDY)  tmp_isr |= 0x10;
    if (m68681->mr1b & 64)  {
        if (m68681->srb & FFULL)  tmp_isr |= 0x20;
    }
    else  {
        if (m68681->sra & RxRDY)  tmp_isr |= 0x20;
    }

    m68681->isr = tmp_isr;
    if (m68681->isr & m68681->imr)
        intSetIRQ(dev);
    else
        intClearIRQ(dev);
}



/*
 * m 6 8 6 8 1 R e s e t
 *
 * Reset the device
 */
int m68681Reset(Device *dev)
{
    M68681  *m68681 = (M68681 *) dev->priv;

    if (!quiet)  uiPrintf(" M68681");

    m68681->ivr = 0x0F;			/* Uninitialized interrupt vector */
    m68681->isr = m68681->imr = 0x00;
    m68681->sra = m68681->srb = 0x00;

    m68681->rxaState = m68681->txaState = INACTIVE;
    m68681->rxbState = m68681->txbState = INACTIVE;

    m68681->mr1aPtr = m68681->mr1bPtr = 1;	/* Pointer to MRx1 */

    return 0;
}



/*
 * m 6 8 6 8 1 R e a d
 *
 * Read one byte
 */
static u_long m68681Read(Device *dev, u_long offset)
{
    M68681  *m68681 = (M68681 *) dev->priv;
    u_char  data;

    switch (offset & (REGSIZE-1))  {
        case MRxA :
		if (m68681->mr1aPtr)  {
			m68681->mr1aPtr = 0;
			data = m68681->mr1a;
		}
		else  {
			data = m68681->mr2a;
		}
		break;

	    case SRA :
		data = m68681->sra;
		break;

	    case RBA :
		m68681->sra &= ~RxRDY;
		m68681->sra &= ~FFULL;
		data = m68681->rba;
		setInterrupt(dev);
		break;

            case ISR :
		data = m68681->isr;
		break;

	    case MRxB :
		if (m68681->mr1bPtr)  {
			m68681->mr1bPtr = 0;
			data = m68681->mr1b;
		}
		else  {
			data = m68681->mr1b;
		}
		break;

	    case SRB :
		data = m68681->srb;
		break;

	    case RBB :
		m68681->srb &= ~RxRDY;
		m68681->srb &= ~FFULL;
		data = m68681->rbb;
		setInterrupt(dev);
		break;

	    case IVR :
		data = m68681->ivr;
		break;

	    default :
		data = 0xff;
		break;
	}


	return (u_long) data;
}



/*
 * m 6 8 6 8 1 W r i t e
 *
 * Write one byte
 */
static void m68681Write(Device *dev, u_long offset, u_char data)
{
	M68681  *m68681 = (M68681 *) dev->priv;

	switch (offset & (REGSIZE-1))  {
	    case MRxA :
		if (m68681->mr1aPtr)  {
			m68681->mr1aPtr = 0;
			m68681->mr1a = data;
			setInterrupt(dev);
		}
		else  {
			m68681->mr2a = data;
		}
		break;

	    case CSRA :
		m68681->csra = data;
		break;

	    case CRA :
		switch ((data & 0x70) >> 4)  {
		    case 1 :
			/* Reset mode register pointer */
			m68681->mr1aPtr = 1;
			break;
		    case 2 :
			/* Reset receiver */
			m68681->rxaState = INACTIVE;
			m68681->sra &= ~RxRDY;
			m68681->sra &= ~FFULL;
			setInterrupt(dev);
			break;
		    case 3 :
			/* Reset transmitter */
			m68681->txaState = INACTIVE;
			m68681->sra &= ~TxRDY;
			m68681->sra &= ~TxEMT;
			setInterrupt(dev);
			break;
		    case 4 :
			/* Clear error status */
			m68681->sra &= 0x0F;
			break;
		}
		switch ((data & 0x0C) >> 2)  {
		    case 1 :
			/* Enable Transmitter */
			m68681->txaState = ACTIVE;
			m68681->sra |= TxRDY;
			m68681->sra |= TxEMT;
			setInterrupt(dev);
			break;
		    case 2 :
			/* Disable Transmitter */
			m68681->txaState = INACTIVE;
			m68681->sra &= ~TxRDY;
			m68681->sra &= ~TxEMT;
			setInterrupt(dev);
			break;
		}
		switch ((data & 0x03) >> 0)  {
		    case 1 :
			/* Enable receiver */
			m68681->rxaState = ACTIVE;
			m68681->sra &= ~RxRDY;
			m68681->sra &= ~FFULL;
			break;
		    case 2 :
			/* Disable receiver */
			m68681->rxaState = INACTIVE;
			break;
		}
		break;

	    case TBA :
		m68681->tba = data;
		m68681->sra &= ~TxRDY;
		m68681->sra &= ~TxEMT;
#ifdef WIN32
#ifdef _CONSOLE
        putchar(data);
		if (data == 0x0a)  putchar(0x0d);
#else
		uiPutchar(data);
#endif
#else
		write(m68681->aFd, &data, 1);
		if (data == 0x0a)  {
			data = 0x0d;
			write(m68681->aFd, &data, 1);
		}
#endif
		m68681->sra |= TxRDY;
		m68681->sra |= TxEMT;
		break;

	    case IMR :
		m68681->imr = data;
		setInterrupt(dev);
		break;

	    case MRxB :
		if (m68681->mr1bPtr)  {
			m68681->mr1bPtr = 0;
			m68681->mr1b = data;
			setInterrupt(dev);
		}
		else  {
			m68681->mr2b = data;
		}
		break;

	    case CSRB :
		m68681->csrb = data;
		break;

	    case CRB :
		switch ((data & 0x70) >> 4)  {
		    case 1 :
			/* Reset mode register pointer */
			m68681->mr1bPtr = 1;
			break;
		    case 2 :
			/* Reset receiver */
			m68681->rxbState = INACTIVE;
			m68681->srb &= ~RxRDY;
			m68681->srb &= ~FFULL;
			setInterrupt(dev);
			break;
		    case 3 :
			/* Reset transmitter */
			m68681->txbState = INACTIVE;
			m68681->srb &= ~TxRDY;
			m68681->srb &= ~TxEMT;
			setInterrupt(dev);
			break;
		    case 4 :
			/* Clear error status */
			m68681->srb &= 0x0F;
			break;
		}
		switch ((data & 0x0C) >> 2)  {
		    case 1 :
			/* Enable Transmitter */
			m68681->txbState = ACTIVE;
			m68681->srb |= TxRDY;
			m68681->srb |= TxEMT;
			setInterrupt(dev);
		    case 2 :
			/* Disable Transmitter */
			m68681->txbState = INACTIVE;
			m68681->srb &= ~TxRDY;
			m68681->srb &= ~TxEMT;
			setInterrupt(dev);
			break;
		}
		switch ((data & 0x03) >> 0)  {
		    case 1 :
			/* Enable receiver */
			m68681->rxbState = ACTIVE;
			m68681->srb &= ~RxRDY;
			m68681->srb &= ~FFULL;
			setInterrupt(dev);
			break;
		    case 2 :
			/* Disable receiver */
			m68681->rxbState = INACTIVE;
			break;
		}
		break;

	    case TBB :
		m68681->tbb = data;
		m68681->srb &= ~TxRDY;
		m68681->srb &= ~TxEMT;
		putchar(data);
		m68681->srb |= TxRDY;
		m68681->srb |= TxEMT;
		break;

	    case IVR :
		m68681->ivr = data;
		break;
	}
}



/*
 * m 6 8 6 8 1 I A c k
 *
 * Interrupt Acknowledge
 */
static u_char m68681IAck(Device *dev, int level)
{
    M68681  *m68681 = (M68681 *) dev->priv;

    fflush(stdout);
    fflush(stderr);

    return m68681->ivr;
}



/*
 * m 6 8 6 8 1 C y c l e
 *
 * Cycle function
 */
static void m68681Cycle(Device *dev)
{
    M68681  *m68681 = (M68681 *) dev->priv;
    u_char data;
#ifdef WIN32
#ifdef _CONSOLE
    if (kbhit())  {
        data = (u_char)getch();
#else
	int  c;

	if ((c = uiGetchar()) >= 0)  {
		data = (u_char)c;
#endif
#else
    fd_set inFdSet;
    struct timeval timeout;

    FD_ZERO(&inFdSet); FD_SET(m68681->aFd, &inFdSet);
    timeout.tv_sec = timeout.tv_usec = 0;
    select(m68681->aFd + 1, &inFdSet, NULL, NULL, &timeout);
    if (FD_ISSET(m68681->aFd, &inFdSet))  {
        read(m68681->aFd, &data, 1);
#endif
        m68681->rba = data;
        m68681->sra |= (RxRDY | FFULL);
        setInterrupt(dev);
    }
}



/*
 * i n i t M 6 8 6 8 1
 *
 * Initialize the device
 */
int initM68681(Device *dev, u_long base, u_long size, void *arg)
{
    M68681 *m68681;
    char   *devName = (char *) arg;
#ifndef WIN32
    int    devFd;
    char   msgBuffer[128];
#endif

    if ((m68681 = malloc(sizeof(M68681))) == NULL) {
        uiPrintf("M68681: Malloc of %ld bytes failed - %s\n", size, strerror(errno));
        return(errno);
    }

#ifndef WIN32
    /* Open and initialize device used for port A */
    if ((devFd = open(devName, O_RDWR | O_NDELAY)) <= 0)  {
        uiPrintf("M68681: Can't open %s - %s\n", devName, strerror(errno));
        return(errno);
    }
    tcgetattr(devFd, &idevSettings);
    devSettings = idevSettings;
#ifdef __sun
    devSettings.c_lflag &= ~(ICANON | IEXTEN | ISIG | ECHO);
    devSettings.c_iflag &= ~(ICRNL | INPCK | ISTRIP | IXON | BRKINT);
    devSettings.c_oflag &= ~OPOST;
    devSettings.c_cflag |= CS8;
    devSettings.c_cc[VMIN]  = 1;
    devSettings.c_cc[VTIME] = 0;
#else
    cfmakeraw(&devSettings);
#endif
    devSettings.c_oflag &= ~ONLCR;
    tcsetattr(devFd, TCSANOW, &devSettings);

    sprintf(msgBuffer, "M68681 Port A ...\n\r");
    write(devFd, msgBuffer, strlen(msgBuffer));

    m68681->aFd = devFd;
#endif

    /* Setup device structure */
    dev->base = (u_long) base;
    dev->mask = ~(size-1);
    dev->priv = (void *) m68681;
    dev->devReset = m68681Reset;
    dev->devRead  = m68681Read;
    dev->devWrite = m68681Write;
    dev->devIAck  = m68681IAck;
    dev->devCycle = m68681Cycle;

    return 0;
}
